package edu.ucr.benchmark;

/*
 * Copyright 2020 Renjie Wu
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import com.dtw.FastDTW;
import com.timeseries.TimeSeries;
import edu.ucr.impl.DtwUpd;
import edu.ucr.util.ProgressBar;
import edu.ucr.util.SquaredDistance;
import edu.ucr.util.TsvReader;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;

public class WarpWindowPairwiseBenchmark {
    private static final int MEASURE_ROUNDS = 10;

    public static void main(String[] args) throws IOException {
        var label = new ArrayList<Integer>();
        var data = new ArrayList<double[]>();
        var dataTS = new ArrayList<TimeSeries>();

        var dtwMethod = Integer.parseInt(args[1]);
        var start = Integer.parseInt(args[2]);
        var end = Integer.parseInt(args[3]);
        var step = Integer.parseInt(args[4]);

        TsvReader.readTsv(args[0], data, dataTS, label);

        System.out.format("DTW method: %d (0: FastDTW, 1: cDTW)\n", dtwMethod);
        try (var outputWriter = new PrintWriter(args[5])) {
            outputWriter.print("warpWindowSize");
            for (int i = 0; i < MEASURE_ROUNDS; i++) {
                outputWriter.print(",time" + (i + 1));
            }
            outputWriter.print(",averageTime");
            outputWriter.println();

            // warp window
            for (int i = start; i <= end; i += step) {
                var warpWindowSize = (int) Math.ceil(data.get(0).length * i / 100.0);
                if (dtwMethod == 0) {
                    warpWindowSize = i; // Use exact number for FastDTW's radius instead of percentage
                }
                System.out.println("-----------------------------------------------");
                System.out.format("Warp window size: %d", warpWindowSize);
                if (dtwMethod == 1) {
                    System.out.format(" (%d%%)", i);
                }
                System.out.println();
                outputWriter.print(warpWindowSize);

                long totalElapsed = 0;

                int actualMeasureRounds = 0;
                int profiledMeasureRound = MEASURE_ROUNDS;

                for (int k = 0; k < MEASURE_ROUNDS && k < profiledMeasureRound; k++, actualMeasureRounds++) {
                    long roundElapsed = 0;
                    var pairsDone = 0;

                    for (int l = 0; l < label.size(); l++) {
                        for (int m = l + 1; m < label.size(); m++) {
                            long startTime, elapsed;

                            switch (dtwMethod) {
                                default:
                                case 0: // FastDTW
                                    startTime = System.nanoTime();
                                    FastDTW.getWarpDistBetween(dataTS.get(l), dataTS.get(m), warpWindowSize, new SquaredDistance());
                                    elapsed = System.nanoTime() - startTime;
                                    break;

                                case 1: // cDTW
                                    startTime = System.nanoTime();
                                    DtwUpd.computeDtw(data.get(l), data.get(m), warpWindowSize);
                                    elapsed = System.nanoTime() - startTime;
                                    break;
                            }

                            roundElapsed += elapsed;
                            pairsDone++;
                            ProgressBar.show(k + 1, pairsDone, label.size() * (label.size() - 1) / 2);
                        }
                    }

                    totalElapsed += roundElapsed;
                    System.out.format(", time elapsed: %.6fs\n", roundElapsed * 1.0 / 1e9);
                    outputWriter.format(",%.6f", roundElapsed * 1.0 / 1e9);

                    var currentAverage = totalElapsed * 1.0 / actualMeasureRounds / 1e9;
                    if (currentAverage < 3600) {
                        profiledMeasureRound = Math.max(3, (int) Math.ceil(3600.0 / currentAverage)); // 3 times at least
                    } else {
                        profiledMeasureRound = 2; // try only 2 times if average is higher than 1 hr.
                    }
                }

                for (int k = actualMeasureRounds; k < MEASURE_ROUNDS; k++) {
                    outputWriter.write(",N/A");
                }

                System.out.format("Average time: %.6fs\n", totalElapsed * 1.0 / actualMeasureRounds / 1e9);
                outputWriter.format(",%.6f\n", totalElapsed * 1.0 / actualMeasureRounds / 1e9);
                outputWriter.flush();
            }
        }
    }
}